/* WhySynth DSSI software synthesizer GUI
 *
 * Copyright (C) 2008 Sean Bolton
 *
 * Parts of this code come from GTK+, both the library source and
 * the example programs.  Other bits come from gAlan 0.2.0,
 * copyright (C) 1999 Tony Garnock-Jones.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301 USA.
 */

#define _ISOC99_SOURCE
#define _BSD_SOURCE

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <gtk/gtkmain.h>
#include <gtk/gtksignal.h>

#include "gtkknob.h"

#define KNOB_SIZE               32
#define HALF_KNOB_SIZE          (KNOB_SIZE / 2)

#define UPDATE_DELAY_LENGTH    300


enum {
    STATE_IDLE,
    STATE_PRESSED,
    STATE_DRAGGING
};

static GtkWidgetClass *parent_class = NULL;

static struct {
    int refcount;
    int initialized;
    cairo_surface_t *knob_surface;
    cairo_surface_t *insensitive_knob_surface;
    cairo_pattern_t *shadow_pattern;
} common;

G_DEFINE_TYPE (GtkKnob, gtk_knob, GTK_TYPE_WIDGET);

static void     gtk_knob_class_init    (GtkKnobClass *klass);
static void     gtk_knob_init          (GtkKnob *knob);
static void     gtk_knob_destroy       (GtkObject *object);
static void     gtk_knob_finalize      (GObject   *object);
static void     gtk_knob_realize       (GtkWidget *widget);
static void     gtk_knob_size_request  (GtkWidget *widget, GtkRequisition *requisition);
static void     gtk_knob_size_allocate (GtkWidget *widget, GtkAllocation  *allocation);
static gboolean gtk_knob_expose        (GtkWidget *knob, GdkEventExpose *event);
static gint     gtk_knob_button_press  (GtkWidget *widget, GdkEventButton *event);
static gint     gtk_knob_button_release(GtkWidget *widget, GdkEventButton *event);
static gint     gtk_knob_motion_notify (GtkWidget *widget, GdkEventMotion *event);
static gint     gtk_knob_timer_callback(GtkKnob *knob);
static void     gtk_knob_update_mouse  (GtkKnob *knob, gint x, gint y, gboolean absolute);
static void     gtk_knob_adjustment_value_changed (GtkAdjustment *adjustment,
                                                   gpointer       data);
static void     gtk_knob_adjustment_changed (GtkAdjustment *adjustment,
                                             gpointer       data);
static void     gtk_knob_common_initialize(void);
static void     gtk_knob_common_finalize(void);
static void     gtk_knob_paint(GtkWidget *widget, cairo_t *cr);


static void
gtk_knob_class_init (GtkKnobClass *class)
{
    GObjectClass   *gobject_class = G_OBJECT_CLASS (class);
    GtkObjectClass *object_class = (GtkObjectClass*) class;
    GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (class);

    parent_class = g_type_class_peek_parent (class);

    gobject_class->finalize = gtk_knob_finalize;
    object_class->destroy = gtk_knob_destroy;

    widget_class->realize = gtk_knob_realize;
    widget_class->size_request = gtk_knob_size_request;
    widget_class->size_allocate = gtk_knob_size_allocate;
    widget_class->expose_event = gtk_knob_expose;
    widget_class->button_press_event = gtk_knob_button_press;
    widget_class->button_release_event = gtk_knob_button_release;
    widget_class->motion_notify_event = gtk_knob_motion_notify;

    common.refcount = 0;
    common.initialized = FALSE;
}


static void
gtk_knob_init (GtkKnob *knob)
{
    knob->adjustment = NULL;
    knob->policy = GTK_UPDATE_CONTINUOUS;
    knob->state = STATE_IDLE;
    knob->center_x = 0;
    knob->center_y = 0;
    knob->saved_x = 0;
    knob->saved_y = 0;
    knob->old_value = 0.0;
    knob->timer = 0;

    common.refcount++;
}


GtkWidget*
gtk_knob_new (GtkAdjustment *adjustment)
{
    GtkKnob *knob = g_object_new (GTK_TYPE_KNOB, NULL);

    gtk_knob_set_adjustment(knob, adjustment);

    return GTK_WIDGET(knob);
}


static void
gtk_knob_finalize (GObject *object)
{
    /* GtkKnob *knob = GTK_KNOB (object); */

    common.refcount--;
    if (common.refcount == 0) gtk_knob_common_finalize();

    (* G_OBJECT_CLASS (parent_class)->finalize) (object);
}


static void
gtk_knob_destroy (GtkObject *object)
{
    GtkKnob *knob = GTK_KNOB(object);

    if (knob->adjustment) {
        g_signal_handlers_disconnect_by_func (knob->adjustment,
                                              gtk_knob_adjustment_changed,
                                              knob);
        g_signal_handlers_disconnect_by_func (knob->adjustment,
                                              gtk_knob_adjustment_value_changed,
                                              knob);
        gtk_object_unref(GTK_OBJECT(knob->adjustment));
        knob->adjustment = NULL;
    }
    if (knob->timer) {
        gtk_timeout_remove(knob->timer);
        knob->timer = 0;
    }

    (* GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}


GtkAdjustment*
gtk_knob_get_adjustment (GtkKnob *knob)
{
    g_return_val_if_fail (GTK_IS_KNOB (knob), NULL);

    if (!knob->adjustment)
        gtk_knob_set_adjustment (knob, NULL);

    return knob->adjustment;
}


void
gtk_knob_set_adjustment (GtkKnob *knob, GtkAdjustment *adjustment)
{
    g_return_if_fail (GTK_IS_KNOB (knob));

    if (!adjustment)
        adjustment = (GtkAdjustment*) gtk_adjustment_new (0.0, 0.0, 0.0, 0.0, 0.0, 0.0);
    else
        g_return_if_fail (GTK_IS_ADJUSTMENT (adjustment));

    if (knob->adjustment) {
        g_signal_handlers_disconnect_by_func (knob->adjustment,
                                              gtk_knob_adjustment_changed,
                                              knob);
        g_signal_handlers_disconnect_by_func (knob->adjustment,
                                              gtk_knob_adjustment_value_changed,
                                              knob);
        gtk_object_unref(GTK_OBJECT(knob->adjustment));
    }

    knob->adjustment = adjustment;
    gtk_object_ref(GTK_OBJECT(knob->adjustment));
    gtk_object_sink(GTK_OBJECT(knob->adjustment));

    gtk_signal_connect(GTK_OBJECT(adjustment), "changed",
                       GTK_SIGNAL_FUNC(gtk_knob_adjustment_changed), (gpointer) knob);
    gtk_signal_connect(GTK_OBJECT(adjustment), "value_changed",
                       GTK_SIGNAL_FUNC(gtk_knob_adjustment_value_changed), (gpointer) knob);

    gtk_knob_adjustment_changed (adjustment, knob);
}


void
gtk_knob_set_update_policy(GtkKnob *knob, GtkUpdateType policy)
{
    g_return_if_fail (GTK_IS_KNOB (knob));

    knob->policy = policy;
}


static void
gtk_knob_realize (GtkWidget *widget)
{
    GtkKnob *knob;
    GdkWindowAttr attributes;
    gint attributes_mask;

    g_return_if_fail (GTK_IS_KNOB (widget));

    knob = GTK_KNOB (widget);
    GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);

    attributes.window_type = GDK_WINDOW_CHILD;
    attributes.x = widget->allocation.x;
    attributes.y = widget->allocation.y;
    attributes.width = widget->allocation.width;
    attributes.height = widget->allocation.height;
    attributes.wclass = GDK_INPUT_OUTPUT;
    attributes.visual = gtk_widget_get_visual (widget);
    attributes.colormap = gtk_widget_get_colormap (widget);
    attributes.event_mask = gtk_widget_get_events (widget) | GDK_EXPOSURE_MASK |
                                GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
                                GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK;

    attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;

    widget->window = gdk_window_new (gtk_widget_get_parent_window (widget), &attributes, attributes_mask);
    gdk_window_set_user_data (widget->window, knob);

    widget->style = gtk_style_attach (widget->style, widget->window);
    gtk_style_set_background (widget->style, widget->window, GTK_STATE_NORMAL);

    gtk_knob_common_initialize();
}


static void
gtk_knob_size_request (GtkWidget *widget, GtkRequisition *requisition)
{
    requisition->width = KNOB_SIZE;
    requisition->height = KNOB_SIZE;
}


static void
gtk_knob_size_allocate (GtkWidget     *widget,
                        GtkAllocation *allocation)
{
    g_return_if_fail (GTK_IS_KNOB (widget));
    g_return_if_fail (allocation != NULL);

    widget->allocation = *allocation;
 
    if (GTK_WIDGET_REALIZED (widget)) {
        gdk_window_move_resize (widget->window,
                                allocation->x, allocation->y,
                                allocation->width, allocation->height);
    }
}


static gboolean
gtk_knob_expose (GtkWidget *widget, GdkEventExpose *event)
{
    int center_x, center_y;
    int l, t, r, b;
    cairo_t *cr;

    /* minimize the clip area to just what we'll actually draw (32x32 at most) */
    center_x = widget->allocation.width / 2;
    center_y = widget->allocation.height / 2;
    if ((center_x - HALF_KNOB_SIZE >= event->area.x + event->area.width)  ||
        (center_x + HALF_KNOB_SIZE <= event->area.x)                      ||
        (center_y - HALF_KNOB_SIZE >= event->area.y + event->area.height) ||
        (center_y + HALF_KNOB_SIZE <= event->area.y)) {
        /* clip area is empty */
        return FALSE;
    }
    l = MAX(event->area.x,                      center_x - HALF_KNOB_SIZE);
    r = MIN(event->area.x + event->area.width,  center_x + HALF_KNOB_SIZE);
    t = MAX(event->area.y,                      center_y - HALF_KNOB_SIZE);
    b = MIN(event->area.y + event->area.height, center_y + HALF_KNOB_SIZE);
    cr = gdk_cairo_create (widget->window);
    cairo_rectangle (cr, l, t, r - l, b - t);
    cairo_clip (cr);
    cairo_translate(cr, center_x - HALF_KNOB_SIZE, center_y - HALF_KNOB_SIZE);

    gtk_knob_paint(widget, cr);

    cairo_destroy (cr);

    return FALSE;
}


static gint
gtk_knob_button_press (GtkWidget *widget, GdkEventButton *event)
{
    GtkKnob *knob;

    g_return_val_if_fail(GTK_IS_KNOB(widget), FALSE);
    g_return_val_if_fail(event != NULL, FALSE);

    knob = GTK_KNOB(widget);

    switch (knob->state) {
      case STATE_IDLE:
        switch (event->button) {
          case 1:
          case 3:
            gtk_grab_add(widget);
            knob->state = STATE_PRESSED;
            knob->center_x = widget->allocation.width / 2;
            knob->center_y = widget->allocation.height / 2;
            knob->saved_x = event->x;
            knob->saved_y = event->y;
            break;

          default:
            break;
        }
        break;

      default:
        break;
    }

    return FALSE;
}


static gint
gtk_knob_button_release (GtkWidget *widget, GdkEventButton *event)
{
    GtkKnob *knob;
    gfloat value;

    g_return_val_if_fail(GTK_IS_KNOB(widget), FALSE);
    g_return_val_if_fail(event != NULL, FALSE);

    knob = GTK_KNOB(widget);

    switch (knob->state) {
      case STATE_PRESSED:
        gtk_grab_remove(widget);
        knob->state = STATE_IDLE;

        switch (event->button) {
          case 1:
            value = knob->adjustment->value - knob->adjustment->page_increment;
            value = MAX(MIN(value, knob->adjustment->upper),
                        knob->adjustment->lower);
            if (knob->adjustment->value != value) {
                knob->adjustment->value = value;
                gtk_signal_emit_by_name(GTK_OBJECT(knob->adjustment), "value_changed");
            }
            break;

          case 3:
            value = knob->adjustment->value + knob->adjustment->page_increment;
            value = MAX(MIN(value, knob->adjustment->upper),
                        knob->adjustment->lower);
            if (knob->adjustment->value != value) {
                knob->adjustment->value = value;
                gtk_signal_emit_by_name(GTK_OBJECT(knob->adjustment), "value_changed");
            }
            break;

          default:
            break;
        }
        break;

      case STATE_DRAGGING:
        gtk_grab_remove(widget);
        knob->state = STATE_IDLE;

        if (knob->policy != GTK_UPDATE_CONTINUOUS && knob->old_value != knob->adjustment->value)
            gtk_signal_emit_by_name(GTK_OBJECT(knob->adjustment), "value_changed");

        break;

      default:
        break;
    }

    return FALSE;
}


static gint
gtk_knob_motion_notify (GtkWidget *widget, GdkEventMotion *event)
{
    GtkKnob *knob;
    GdkModifierType mods;
    gint x, y;

    g_return_val_if_fail(GTK_IS_KNOB(widget), FALSE);
    g_return_val_if_fail(event != NULL, FALSE);

    knob = GTK_KNOB(widget);

    x = event->x;
    y = event->y;

    if (event->is_hint || (event->window != widget->window))
        gdk_window_get_pointer(widget->window, &x, &y, &mods);

    switch (knob->state) {
      case STATE_PRESSED:
        knob->state = STATE_DRAGGING;
        /* fall through */

      case STATE_DRAGGING:
        if (mods & GDK_BUTTON1_MASK) {
            gtk_knob_update_mouse(knob, x, y, TRUE);
            return TRUE;
        } else if (mods & GDK_BUTTON3_MASK) {
            gtk_knob_update_mouse(knob, x, y, FALSE);
            return TRUE;
        }
        break;

      default:
        break;
    }

    return FALSE;
}


static gint
gtk_knob_timer_callback (GtkKnob *knob)
{
    g_return_val_if_fail(GTK_IS_KNOB(knob), FALSE);

    if (knob->policy == GTK_UPDATE_DELAYED)
        gtk_signal_emit_by_name(GTK_OBJECT(knob->adjustment), "value_changed");

    return FALSE;       /* don't keep running this timer */
}


static void
gtk_knob_update_mouse(GtkKnob *knob, gint x, gint y, gboolean absolute)
{
    gfloat old_value, new_value;
    gdouble angle;
    gint dv, dh;

    g_return_if_fail(GTK_IS_KNOB(knob));

    old_value = knob->adjustment->value;

    if (absolute) {

        angle = atan2(-y + knob->center_y, x - knob->center_x);
        angle /= M_PI;
        if (angle < -0.5)
            angle += 2;

        new_value = -(2.0/3.0) * (angle - 1.25);   /* map [1.25pi, -0.25pi] onto [0, 1] */
        new_value *= knob->adjustment->upper - knob->adjustment->lower;
        new_value += knob->adjustment->lower;

    } else {

        dv = knob->center_y - y; /* inverted cartesian graphics coordinate system */
        dh = x - knob->center_x;

        if (abs(dv) < HALF_KNOB_SIZE && abs(dh) < HALF_KNOB_SIZE) {
            
            /* translate close-in motion into fine adjustment */
            dv = knob->saved_y - y;
            dh = x - knob->saved_x;
            new_value = knob->adjustment->value +
                        (float)(dv + dh) * knob->adjustment->step_increment;

        } else if (abs(dv) > abs(dh)) {

            /* translate up-and-down motion into fine adjustment */
            dv = knob->saved_y - y;
            new_value = knob->adjustment->value +
                        (float)dv * knob->adjustment->step_increment;

        } else {

            /* translate side-to-side motion into coarse adjustment */
            dh = x - knob->saved_x;
            new_value = knob->adjustment->value +
                        (float)dh * (knob->adjustment->upper -
                                     knob->adjustment->lower) / 300.0f;

        }
        knob->saved_x = x;
        knob->saved_y = y;
    }

    new_value = MAX(MIN(new_value, knob->adjustment->upper),
                    knob->adjustment->lower);

    knob->adjustment->value = new_value;

    if (knob->adjustment->value != old_value) {

        if (knob->policy == GTK_UPDATE_CONTINUOUS)
            gtk_signal_emit_by_name(GTK_OBJECT(knob->adjustment), "value_changed");
        else {
            gtk_widget_queue_draw (GTK_WIDGET(knob));

            if (knob->policy == GTK_UPDATE_DELAYED) {
                if (knob->timer)
                    gtk_timeout_remove(knob->timer);

                knob->timer = gtk_timeout_add (UPDATE_DELAY_LENGTH,
                                               (GtkFunction) gtk_knob_timer_callback,
                                               (gpointer) knob);
          }
        }
    }
}


static void
gtk_knob_adjustment_changed(GtkAdjustment *adjustment, gpointer data)
{
    GtkKnob *knob;

    g_return_if_fail(adjustment != NULL);
    g_return_if_fail(data != NULL);

    knob = GTK_KNOB(data);

    knob->old_value = adjustment->value;

    gtk_widget_queue_draw (GTK_WIDGET(knob));
}

static void
gtk_knob_adjustment_value_changed (GtkAdjustment *adjustment, gpointer data)
{
    GtkKnob *knob;

    g_return_if_fail(adjustment != NULL);
    g_return_if_fail(data != NULL);

    knob = GTK_KNOB(data);

    if (knob->old_value != adjustment->value) {
        knob->old_value = adjustment->value;

        gtk_widget_queue_draw (GTK_WIDGET(knob));
    }
}


/* forward declarations of PNG data below */
static size_t        knob_c_png_length;
static unsigned char knob_c_png[];
static size_t        knob_i_png_length;
static unsigned char knob_i_png[];

struct cairo_read_func_closure {
    size_t         length;
    size_t         offset;
    unsigned char *data;
};

static cairo_status_t
cairo_read_func(void *closure, unsigned char *data, unsigned int length)
{
    struct cairo_read_func_closure *cl = (struct cairo_read_func_closure *)closure;
    size_t l = MIN(length, cl->length - cl->offset);

    memcpy(data, cl->data + cl->offset, l);

    cl->offset += l;

    return CAIRO_STATUS_SUCCESS;
}

static void
gtk_knob_common_initialize(void)
{
    struct cairo_read_func_closure cl;
    cairo_pattern_t *pat;

    if (common.initialized)
        return;

    /* create knob surfaces */
    cl.length = knob_c_png_length;
    cl.offset = 0;
    cl.data = knob_c_png;
    common.knob_surface = cairo_image_surface_create_from_png_stream(cairo_read_func, &cl);
    cl.length = knob_i_png_length;
    cl.offset = 0;
    cl.data = knob_i_png;
    common.insensitive_knob_surface = cairo_image_surface_create_from_png_stream(cairo_read_func, &cl);

    /* create the shadow pattern */
    pat = cairo_pattern_create_radial (KNOB_SIZE * 0.61, KNOB_SIZE * 0.61, 0.0,
                                       KNOB_SIZE * 0.61, KNOB_SIZE * 0.61, KNOB_SIZE * 0.4);  /* 0.375 */
    cairo_pattern_add_color_stop_rgba (pat, 0.75, 0, 0, 0, 0.5);
    cairo_pattern_add_color_stop_rgba (pat, 0.90, 0, 0, 0, 0.1);
    cairo_pattern_add_color_stop_rgba (pat, 1.00, 0, 0, 0, 0);
    common.shadow_pattern = pat;
    
    common.initialized = TRUE;
}

static void
gtk_knob_common_finalize(void)
{
    if (!common.initialized)
        return;

    cairo_surface_destroy (common.knob_surface);
    cairo_surface_destroy (common.insensitive_knob_surface);
    cairo_pattern_destroy (common.shadow_pattern);
    
    common.initialized = FALSE;
}

static void
gtk_knob_paint(GtkWidget *widget, cairo_t *cr)
{
    GtkKnob *knob = GTK_KNOB(widget);
    gfloat value, range;

    if (GTK_WIDGET_IS_SENSITIVE (widget)) {

        range = knob->adjustment->upper - knob->adjustment->lower;
        if (fabs(range) < 1e-37) {
            value = 0.0;
        } else {
            value = MAX(MIN(knob->adjustment->value, knob->adjustment->upper),
                        knob->adjustment->lower);
            value = value - knob->adjustment->lower;
            value /= range;  /* now scaled 0.0 to 1.0 */
        }

        /* draw the red arc */
        cairo_set_source_rgb (cr, 1.0, 0, 0);
        cairo_move_to (cr, HALF_KNOB_SIZE, HALF_KNOB_SIZE);
        cairo_arc (cr, HALF_KNOB_SIZE, HALF_KNOB_SIZE, (KNOB_SIZE * 0.47),
                   M_PI * 0.75, M_PI * 0.75 + value * (1.5 * M_PI));
        cairo_close_path (cr);
        cairo_fill (cr);

        /* draw the knob shadow */
        cairo_set_source (cr, common.shadow_pattern);
        cairo_paint (cr);

        /* paint the knob and reference points over what we have so far */
        cairo_set_source_surface (cr, common.knob_surface, 0, 0);
        cairo_paint (cr);

        /* add cursor to knob */
        cairo_translate (cr, HALF_KNOB_SIZE, HALF_KNOB_SIZE);
        cairo_rotate (cr, M_PI * 0.75 + value * (1.5 * M_PI));
        cairo_set_line_width (cr, KNOB_SIZE / 32.0 * 1.5);
        cairo_set_source_rgb (cr, 1.0, 1.0, 1.0);
        cairo_move_to (cr, KNOB_SIZE * 0.06, 0);
        cairo_line_to (cr, KNOB_SIZE * 0.30, 0);
        cairo_stroke (cr);

    } else {  /* insensitive */
        cairo_set_source_surface (cr, common.insensitive_knob_surface, 0, 0);
        cairo_paint (cr);
    }
}

static size_t        knob_c_png_length = 1018;
static unsigned char knob_c_png[1018] =
  "\211PNG\15\12\32\12\0\0\0\15IHDR\0\0\0\40\0\0\0\40\10\6\0\0\0szz\364"
  "\0\0\0\1sRGB\0\256\316\34\351\0\0\3\264IDATX\303\355\226=o\33G\20\206"
  "\337;\362\370\5\27\376\3\2l\3\252\\P\255\33\265\6\201TA\212\224\1\235"
  "*M:\247O~B`\40\205\341V\277\200`g\300.RK*\2434\1l\230\242\216\342\361"
  ">vwvggS\34I\310\222\22\221tR\4\361\0[\355\307<;\363\316\354\2\377q\13"
  "\313\261\263\305\370\277[\264\313\246\243\243_\207J\331\201\326\274o"
  "\255\203H8\23\221\361\363\347_\274\372\327\0F\243\323C\";f\226\36\263"
  "\207R\4c\34\254e\210\4x\357\1\0JY\225\246\363\301\213\27\337\276\375"
  "\307\0F\243\343c\40\352;\347a-#\317\25\362\\\3039\2061\356#\0\"\7\245"
  "\10\227\227\305\311\321\321\367\7\237\14""0\32\235\254U\276X(dY\205\331"
  "\254\200\326\26Z[\204\20\340\34\177\264\247\236#\234\237\317\360\372"
  "\365\217\321\316\0\327\235\247i\201\3514C\236+\30c\341\\}k\21\271\261"
  "\227\310\341\3622\307dr\201\323\323\237\243\255\313\260\16{mUEH\323\2"
  "\223\311\34\263Y\201\305\242\2\221\203\210\334\352\34\0\32\215\30\255"
  "V\23\235N\33\217\36\15\217\267\2\30\215N\17\253\312\366\211\34\26\13"
  "\205\311$\303t\232!M\27P\312l$\332\32\40A\267\333A\253\325\352\77|\370"
  "\354pc\200,\253\306\0P\226f\35\366\3514\3\221\333\274\274\242\10\255"
  "V\202f\263\201$iB\304\2157\6\210\"\364\230=\312\322`6+0\237\227[9\277"
  "\32\205n\267\215V+A\24\305\275\215\0^\276|3d\256\313M)\13\245\14\224"
  "\242\235\272\234H\255a\357\5\336\33\354\355}=\274\23\40\212\302@$\200"
  "\331CkZ\213m[\253/\341@\344\340\275G\10\14\346rp}]\363\306\363\26\260"
  "\277\212\0\263\300\373\355\235\207\20\226\315I\226\21\324`.\341\275\336"
  "\277\23`\25:cj\362\353Mf\23\363^\20B\200\265\16\326Z\20\25`V\20\241\273"
  "5\220e\352,\3175\252\312\200\310\255\233\315\266\271\267\326\255\323"
  "\340\275\202\210\1\200\263;#\220\246\213q\10\370\252~Xh\353\3743\373"
  "u\12\210\354\32\200Y!\4\2147j\305\303\341/\241\16e}\330\252\2567\5\320"
  "\232P\226\32Y\226\343\374|\212\262\374\35D)\312\362\267h#\15""0\263r"
  "\316\367\254\255k\277\323i\325\213\233\215\33b\273*\322\325\255\313R"
  "\243(*\314fsh\375\16\316-\40b\325m\276\376\12`\240\265}\263\2\360^n@"
  "\254\234{/p\216\327\302s\216\221e9f\2639\212\342\17\30""3\5s\11\221x"
  "\260\325k\370\364\351O\307\326\332\276\367\202$i\342\336\275\36\222\244"
  "\211v;\271!4c,\230\31u\367\2540\237/PU\357\241\365;0\27\10\301\235h\375"
  "\341`\353\347\370\311\223\37\202\265\26\215F\3I\222\240\323i\241\333"
  "m\243\321h\300{\277\16\271R\6Z\33\30CP*\2071\23\20]\300\3329D\34\210"
  "\316\243\235\77$\217\37\177\27\3428\206\210\40\216cDQ\204$i\"\212\"0"
  "{8\347`\255\2031\5\230\13\20\245p.\7s\1\21\2\321\305N\37\222pu\376\301"
  "\203o\216\231\313\276\367\6@@\10\36q\334^\246\202\340\275\206s\5\274"
  "W\313A\0\344\204\350\342\340\266\363\266\6\0\200\275\275/\17\253\352"
  "\375\330{\323\273Z\222\336\23Dh\11B\0D\305q{`\314\207\267\177w\336'}"
  "\313\357\337\357\17\201\346\200\271\330\27!0\227g!\304c\347.^\341\263"
  "}\266-\355O\3\341\3\357[\342\211\34\0\0\0\0IEND\256B`\202";

static size_t        knob_i_png_length = 819;
static unsigned char knob_i_png[819] =
  "\211PNG\15\12\32\12\0\0\0\15IHDR\0\0\0\40\0\0\0\40\10\6\0\0\0szz\364"
  "\0\0\0\1sRGB\0\256\316\34\351\0\0\2\355IDATX\303\355W\273N#A\20\254\331"
  "]\303O8#'u@d\211\4\221\372\3\10\220\34\20\\\200D\300\27\40K\266d\210"
  "\20\374\7!1\21\22\"[9\363\37\354kf\272{\346\22\317j\15\346\3305Gp:Zr"
  "fOWUwW\267\201\1779\306\343\361\343x<~\374\312\33\21\376\367P\333\374"
  "\350\365\365\365\210\210\6\314\334\27\21""8\347\226\336\373\247\203\203"
  "\203\207o\3\220\246\351>3O\234s\273\336{Xk!\"X\1\200\367\36\0@D\246("
  "\212\213\343\343\343\227\277\6\40M\323{\0{!\2411\246\6\300\314\360\336"
  "\3039\7\0\20\21\20\21\312\262\\\214F\243\323/\3H\323\264\356rc\14\252"
  "\252BUU`f\20\321Z\362\20D\4\"B\236\347899\31n\15\340m\362\242(P\24\5"
  "\2141`f\210\10\0\324\3627CDP\226%\262,\303\331\331\331\260\363\30\256"
  "d\7\0Xk\353\344UUAk]K\277)9\0(\245\20\3071\222$\301l6\273\357\4\40M\323"
  "}k\355\36""3\303\30\203<\317k\0\326\332v\6\23EH\222\4;;;\210\343xo6\233"
  "\355\267\6\240\265\236\274e^\24\5\230\271\323\210\305q\\+\341\234\233"
  "\264\6\240\224\252G\255)y\347\31W\12\275^/\0\331m\5\340\371\371\371("
  "\214[\350\346\266\262\277\215\320\37\336{\210\10&\223\311Q\33\5\6\241"
  "\213\231\271n\266\256\341\234[3*\347\34\210h\360\366{\311\6\324\375M"
  "\16\267\15{\347\34\254\265\260\326\202\210\40\"\375O\1""4\3478\260\330"
  "\206}xCD`\255\5""3\2773\254\215%\320Z/\265\326k^\277\15\373&\201\306"
  ";\313O\1\224e\371T\226e-[\327\22\4\226!y\350\243\225u\77}\12\340\360"
  "\360\360\241i\271\333D\230\240P\377\240\300\325\325\325C\253\36p\316"
  "\31\21\331\15\362'IR\273\333G\214\233\322\33c\326v\207\265\26\3169\323"
  "\332\210D\344\"x@`\262i\353\205)a\346\272dD\204\252\252\220e\31\262,"
  "CUU\40\"\0\270h\15`4\32\2750\363\"$7\306\2041\252g:\324\267\371\35\255"
  "5\262,[[\\+\366\213\351t\372\322y\35\337\335\335=23\242(\2527[\257\327"
  "C\24Ek\354\233j5o\206P\377\371|>\334\372\40\271\271\271yTJ\301{\17\245"
  "\24\224R\210\242\10J\251\2651\13\15\27\256\245\240\330\237\222\177\10"
  "\40\334\372\267\267\267C\0\230N\247\367D\264'\"\365\15\20\307\361;\263"
  "\11\335\277j\336\305|>\77\335\364^\347\377\5\347\347\347\247\0~i\255"
  "M\250m\236\347k\237\306\251f\242(\372\25\222\177\313Y~yyy\244\224\32"
  "Xk\373\253%\263\4\360t}}\375\200\237\370\211\216\361\33\353\274\40}\314"
  "\227E/\0\0\0\0IEND\256B`\202";

